import threading
import logging
import io


class CapturingHandler(logging.StreamHandler):
    @property
    def content(self) -> str:
        return self.stream.getvalue()

    def __init__(self):
        super().__init__(stream=io.StringIO())

    def clear(self):
        self.stream.truncate(0)
        self.stream.seek(0)


_test_index = 0
_index_lock = threading.Lock()


class ExtraLogger(logging.Logger):
    def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
        """Replacement for Logger.makeRecord to overwrite fields via ``extra``"""
        try:
            created = extra and extra.pop('created')
        except KeyError:
            created = None
        rv = super().makeRecord(name, level, fn, lno, msg, args, exc_info, func, None, sinfo)
        if extra is not None:
            for key in extra:
                rv.__dict__[key] = extra[key]
        if created:
            rv.created = created
            rv.msecs = (created - int(created)) * 1000
            rv.relativeCreated = (created - logging._startTime) * 1000
        return rv


def make_test_logger(base_name: str = 'test_logger'):
    with _index_lock:
        global _test_index
        log_name = base_name + '.test%d' % _test_index
        _test_index += 1
    logger = logging.getLogger(log_name)
    logger.propagate = False
    logger.__class__ = ExtraLogger
    handler = CapturingHandler()
    logger.handlers = [handler]
    return logger, handler



import threading
import logging
import io

import pytest

from ..mock.pool import FullMockPool

from cobald.decorator.logger import Logger


class CapturingHandler(logging.StreamHandler):
    @property
    def content(self) -> str:
        return self.stream.getvalue()

    def __init__(self):
        super().__init__(stream=io.StringIO())


_test_index = 0
_index_lock = threading.Lock()


def make_logger(base_name: str = 'test_logger'):
    global _test_index
    with _index_lock:
        log_name = base_name + '.test%d' % _test_index
        _test_index += 1
    logger = logging.getLogger(log_name)
    logger.propagate = False
    handler = CapturingHandler()
    logger.handlers = [handler]
    return logger, handler


class TestLogger(object):
    def test_transparent_logging(self):
        pool = FullMockPool()
        logger, handler = make_logger()
        chain = Logger(target=pool, name=logger.name)
        chain.demand += 1
        assert handler.content
        assert chain.demand == pool.demand
        assert chain.utilisation == pool.utilisation
        assert chain.allocation == pool.allocation

    def test_name(self):
        pool = FullMockPool()
        chain = Logger(target=pool, name='default')
        assert chain.name == 'default'
        chain.name = None
        assert chain.name == pool.__class__.__name__
        chain.name = 'final'
        assert chain.name == 'final'



import pytest

from ..mock.pool import FullMockPool

from cobald.decorator.standardiser import Standardiser


class TestStandardiser(object):
    def test_init_enforcement(self):
        pool = FullMockPool()
        with pytest.raises(ValueError):
            Standardiser(pool, minimum=10, maximum=-10)
        with pytest.raises(ValueError):
            Standardiser(pool, surplus=-20)
        with pytest.raises(ValueError):
            Standardiser(pool, backlog=-20)
        with pytest.raises(ValueError):
            Standardiser(pool, granularity=-20)

    def test_granularity(self):
        pool = FullMockPool()
        for granularity in range(1, 5):
            standardiser = Standardiser(pool, granularity=granularity)
            standardiser.demand = 0
            for multiple in range(5):
                for offset in range(granularity):
                    value = granularity * multiple + offset
                    standardiser.demand = value
                    assert pool.demand % granularity == 0

    def test_granularity_incremental(self):
        """Standardiser: ``demand + n`` equals ``demand + 1 + 1 + 1 ... + 1``"""
        pool = FullMockPool()
        for granularity in (1, 3, 5, 7, 13, 16):
            # wrap first, then reset to test if Standardiser correctly picks this up
            standardiser = Standardiser(pool, granularity=granularity)
            pool.demand = 0
            for total in range(granularity * 5):
                standardiser.demand += 1
                assert pool.demand % granularity == 0
                assert standardiser.demand % granularity == (total + 1) % granularity
                assert standardiser.demand == total + 1

    def test_surplus_backlog(self):
        pool = FullMockPool()
        for base_supply in (10000, 500, 0):
            for surplus in (100, 10, 1):
                standardiser = Standardiser(pool, surplus=surplus)
                pool.supply = base_supply
                standardiser.demand = pool.supply + 2 * surplus
                assert pool.demand == base_supply + surplus
                standardiser.demand = pool.supply - 2 * surplus
                assert pool.demand == base_supply - 2 * surplus
            for backlog in (100, 10):
                standardiser = Standardiser(pool, backlog=backlog)
                pool.supply = base_supply
                standardiser.demand = pool.supply + 2 * backlog
                assert pool.demand == base_supply + 2 * backlog
                standardiser.demand = pool.supply - 2 * backlog
                assert pool.demand == base_supply - backlog

    def test_minimum_maximum(self):
        pool = FullMockPool()
        for minimum in (-100, -10, -1, 0, 10, 100):
            standardiser = Standardiser(pool, minimum=minimum)
            standardiser.demand = minimum - abs(minimum)
            assert pool.demand == minimum
            standardiser.demand = minimum + abs(minimum)
            assert pool.demand == minimum + abs(minimum)
        for maximum in (-100, -10, -1, 0, 10, 100):
            standardiser = Standardiser(pool, maximum=maximum)
            standardiser.demand = maximum + abs(maximum)
            assert pool.demand == maximum
            standardiser.demand = maximum - abs(maximum)
            assert pool.demand == maximum - abs(maximum)



import time
import ast

from cobald.monitor.format_line import LineProtocolFormatter

from . import make_test_logger


def parse_line_protocol(literal: str):
    name_tags, _, fields_stamp = literal.strip().partition(' ')
    fields, _, stamp = fields_stamp.partition(' ')
    fields = fields.split(',') if fields else []
    name, *tags = name_tags.split(',')
    return name, {
        key: value
        for key, value
        in (tag.split('=') for tag in tags)
    }, {
        key: ast.literal_eval(value)
        for key, value
        in (field.split('=') for field in fields)
    }, None if not stamp else int(stamp)


class TestFormatLine(object):
    def test_payload(self):
        for payload in (
                {'a': 'a'},
                {str(i): i for i in range(20)},
        ):
            logger, handler = make_test_logger(__name__)
            handler.formatter = LineProtocolFormatter()
            logger.critical('message', payload)
            name, tags, fields, timestamp = parse_line_protocol(handler.content)
            assert len(fields) == len(payload)
            assert fields == payload
            assert timestamp is None

    def test_timestamp_resolution(self):
        now = time.time()
        payload = {'a': 'a', '1': 1, '2.2': 2.2}
        for resolution in (100, 10, 1):
            logger, handler = make_test_logger(__name__)
            handler.formatter = LineProtocolFormatter(resolution=resolution)
            logger.critical('message', payload, extra={'created': now})
            name, tags, fields, timestamp = parse_line_protocol(handler.content)
            assert len(fields) == len(payload)
            assert timestamp == now // resolution * resolution * 1e9

    def test_payload_tag(self):
        for payload in (
                {'a': 'a'},
                {'a': 'a', 'b': 'b'},
                {str(i): i for i in range(20)},
        ):
            logger, handler = make_test_logger(__name__)
            handler.formatter = LineProtocolFormatter({'a'})
            logger.critical('message', payload)
            name, tags, fields, timestamp = parse_line_protocol(handler.content)
            assert timestamp is None
            if 'a' in payload:
                assert len(fields) == len(payload) - 1
                assert tags == {'a': 'a'}
                assert fields == {key: value for key, value in payload.items() if key != 'a'}
            else:
                assert len(fields) == len(payload)
                assert fields == payload

    def test_payload_empty(self):
        logger, handler = make_test_logger(__name__)
        handler.formatter = LineProtocolFormatter()
        logger.critical('message', {})
        name, tags, fields, timestamp = parse_line_protocol(handler.content)
        assert len(fields) == 0
        assert fields == {}
        assert len(tags) == 0
        assert tags == {}
        assert timestamp is None



import json
import logging
import time

from cobald.monitor.format_json import JsonFormatter

from . import make_test_logger


class TestFormatJson(object):
    def test_payload(self):
        for payload in (
                {'a': 'a'},
                {str(i): i for i in range(20)},
        ):
            logger, handler = make_test_logger(__name__)
            handler.formatter = JsonFormatter()
            logger.critical('message', payload)
            data = json.loads(handler.content)
            assert len(data) == len(payload) + 2
            assert data.pop('message') == 'message'
            assert data.pop('time')
            assert data == payload

    def test_default_timestamp(self):
        now = time.time()
        payload = {'a': 'a', '1': 1, '2.2': 2.2}
        logger, handler = make_test_logger(__name__)
        handler.formatter = JsonFormatter()
        logger.critical('message', payload, extra={'created': now})
        data = json.loads(handler.content)
        assert len(data) == len(payload) + 2
        # from Formatter and LogRecord
        ct = time.localtime(now)
        msecs = (now - int(now)) * 1000
        default_time_string = logging.Formatter.default_msec_format % (
            time.strftime(logging.Formatter.default_time_format, ct),
            msecs
        )
        assert data.pop('time') == default_time_string

    def test_explicit_timestamp(self):
        now = time.time()
        payload = {'a': 'a', '1': 1, '2.2': 2.2}
        logger, handler = make_test_logger(__name__)
        handler.formatter = JsonFormatter(datefmt='%Y')
        logger.critical('message', payload, extra={'created': now})
        data = json.loads(handler.content)
        assert len(data) == len(payload) + 2
        # from Formatter and LogRecord
        ct = time.localtime(now)
        default_time_string = time.strftime('%Y', ct)
        assert data.pop('time') == default_time_string

    def test_disabled_timestamp(self):
        now = time.time()
        payload = {'a': 'a', '1': 1, '2.2': 2.2}
        logger, handler = make_test_logger(__name__)
        handler.formatter = JsonFormatter(datefmt='')
        logger.critical('message', payload, extra={'created': now})
        data = json.loads(handler.content)
        assert len(data) == len(payload) + 1
        assert 'time' not in data

    def test_payload_empty(self):
        logger, handler = make_test_logger(__name__)
        handler.formatter = JsonFormatter()
        logger.critical('message', {})
        data = json.loads(handler.content)
        assert len(data) == 2

    def test_default(self):
        logger, handler = make_test_logger(__name__)
        handler.formatter = JsonFormatter(fmt={"test": 1})
        logger.critical('message', {'drones': 1})
        data = json.loads(handler.content)
        assert data.pop("test") == 1
        handler.clear()
        logger.critical('message', {'test': 2})
        data = json.loads(handler.content)
        assert data.pop("test") == 2
        assert len(data) == 2
        handler.clear()
        logger.critical('message', {})
        data = json.loads(handler.content)
        assert data.pop("test") == 1
        assert len(data) == 2



import abc

from ._pool import Pool


class Controller(metaclass=abc.ABCMeta):
    """
    Controller adjusting the demand in a :py:class:`~.Pool`

    :param target: the resource pool for which demand is adjusted
    """
    def __init__(self, target: Pool):
        self.target = target

    @abc.abstractmethod
    def regulate(self, interval: float):
        """Regulate demand for the ``pool`` in the past ``interval`` seconds"""



import pytest

from ..mock.pool import MockPool

from cobald.controller.linear import LinearController
from cobald.controller.switch import DemandSwitch


class TestLinearController(object):
    def test_select(self):
        pool = MockPool()
        switch_controller = DemandSwitch(pool, LinearController(pool, rate=1), 5, LinearController(pool, rate=2))
        pool.utilisation = pool.allocation = 1.0
        expected_demand = 0
        for i in range(5):
            switch_controller.regulate(1)
            expected_demand += 1
            assert(pool.demand == expected_demand)
        for i in range(5):
            assert(pool.demand == expected_demand)
            switch_controller.regulate(1)
            expected_demand += 2
